from typing import TypeVar

from more_properties.util_properties import WrappedProperty

__all__ = [
    "ClassProperty",
    "class_property",
    "StaticProperty",
    "static_property",
]

OT = TypeVar("OT", contravariant=True)  # Owner Type
VT = TypeVar("VT")  # Value Type


class ClassProperty(WrappedProperty[OT, VT]):
    wrapper = classmethod

    def __post_init__(self) -> None:
        fget = self.__dict__["fget"]

        if isinstance(fget, self.wrapper):
            fget.__doc__ = fget.__func__.__doc__

        super().__post_init__()


class StaticProperty(ClassProperty[OT, VT]):
    wrapper = staticmethod


# For some reason, Python 3.6 treats classmethod and staticmethod as abstract methods,
# disallowing instantiation of ClassProperty and StaticProperty
setattr(ClassProperty, "__abstractmethods__", frozenset())
setattr(StaticProperty, "__abstractmethods__", frozenset())

class_property = ClassProperty
static_property = StaticProperty



from dataclasses import dataclass
from typing import Generic, Optional, Type, TypeVar

from more_properties.property import Property
from more_properties.types import Getable

__all__ = [
    "NamedProperty",
    "WrappedProperty",
]

OT = TypeVar("OT", contravariant=True)  # Owner Type
VT = TypeVar("VT")  # Value Type


@dataclass
class NamedProperty(Property[OT, VT]):
    name: Optional[str] = None

    def __set_name__(self, owner: Type[OT], name: str) -> None:
        self.name = name


@dataclass
class TrivialWrapper(Generic[OT, VT]):
    getable: Getable[OT, VT]

    def __get__(self, instance: Optional[OT], owner: Type[OT]) -> VT:
        getable: Getable[OT, VT] = self.__dict__["getable"]
        return getable.__get__(instance, owner)


class WrappedProperty(Property[OT, VT]):
    wrapper: type = TrivialWrapper

    def __post_init__(self) -> None:
        fget = self.__dict__["fget"]

        if fget is not None and not isinstance(fget, self.wrapper):
            setattr(self, "fget", self.wrapper(fget))

        fset = self.__dict__["fset"]

        if fset is not None and not isinstance(fset, self.wrapper):
            setattr(self, "fset", self.wrapper(fset))

        fdel = self.__dict__["fdel"]

        if fdel is not None and not isinstance(fdel, self.wrapper):
            setattr(self, "fdel", self.wrapper(fdel))

        super().__post_init__()



from dataclasses import dataclass, replace
from typing import Generic, Optional, Type, TypeVar

from more_properties.types import Deleter, Getter, Setter

__all__ = [
    "Property",
    "property",
]

OT = TypeVar("OT", contravariant=True)  # Owner Type
VT = TypeVar("VT")  # Value Type


@dataclass
class Property(Generic[OT, VT]):
    fget: Optional[Getter[OT, VT]] = None
    fset: Optional[Setter[OT, VT]] = None
    fdel: Optional[Deleter[OT]] = None
    doc: Optional[str] = None

    def __post_init__(self) -> None:
        self.__doc__ = (
            self.doc
            if self.doc is not None
            else getattr(self.__dict__.get("fget"), "__doc__", None)
        )

    def __get__(self, instance: Optional[OT], owner: Type[OT]) -> VT:
        fget: Optional[Getter[OT, VT]] = self.__dict__["fget"]

        if fget is None:
            raise AttributeError("unreadable attribute")

        return fget.__get__(instance, owner)()

    def __set__(self, instance: OT, value: VT) -> None:
        fset = self.__dict__["fset"]

        if fset is None:
            raise AttributeError("can't set attribute")

        fset.__get__(instance, type(instance))(value)

    def __delete__(self, instance: OT) -> None:
        fdel = self.__dict__["fdel"]

        if fdel is None:
            raise AttributeError("can't delete attribute")

        fdel.__get__(instance, type(instance))()

    def getter(self, func: Getter[OT, VT]) -> "Property[OT, VT]":
        return replace(self, fget=func)

    def setter(self, func: Setter[OT, VT]) -> "Property[OT, VT]":
        return replace(self, fset=func)

    def deleter(self, func: Deleter[OT]) -> "Property[OT, VT]":
        return replace(self, fdel=func)


property = Property



from dataclasses import dataclass
from typing import Optional, Type, TypeVar, Union

from more_properties.class_property import ClassProperty, StaticProperty
from more_properties.util_properties import NamedProperty

__all__ = [
    "cached_property",
    "cached_class_property",
    "cached_static_property",
]

OT = TypeVar("OT", contravariant=True)  # Owner Type
VT = TypeVar("VT")  # Value Type


class Uncached:
    pass


Cache = Union[VT, Uncached]


@dataclass
class CachedProperty(NamedProperty[OT, VT]):
    @property
    def cache_name(self) -> str:
        if self.name is None:
            raise AttributeError(f"Property {self!r} not assigned to class")

        return f"__{self.name}_cache"

    def __get__(self, instance: Optional[OT], owner: Type[OT]) -> VT:
        cache_name = self.cache_name

        if cache_name not in instance.__dict__:
            instance.__dict__[cache_name] = super().__get__(instance, owner)

        value: VT = instance.__dict__[cache_name]

        return value


@dataclass
class CachedClassProperty(ClassProperty[OT, VT], NamedProperty[OT, VT]):
    @property
    def cache_name(self) -> str:
        if self.name is None:
            raise AttributeError(f"Property {self!r} not assigned to class")

        return f"__{self.name}_cache"

    def __get__(self, instance: Optional[OT], owner: Type[OT]) -> VT:
        cache_name = self.cache_name

        if cache_name in owner.__dict__:
            value: VT = owner.__dict__[cache_name]

            return value

        value = super().__get__(instance, owner)

        setattr(owner, cache_name, value)

        return value

    def __set__(self, instance: OT, value: VT) -> None:
        owner = type(instance)
        cache_name = self.cache_name

        if cache_name in owner.__dict__:
            delattr(owner, cache_name)

        super().__set__(instance, value)

    def __delete__(self, instance: OT) -> None:
        owner = type(instance)
        cache_name = self.cache_name

        if cache_name in owner.__dict__:
            delattr(owner, cache_name)

        super().__delete__(instance)


@dataclass
class CachedStaticProperty(StaticProperty[OT, VT]):
    value: Cache[VT] = Uncached()

    def __get__(self, instance: Optional[OT], owner: Type[OT]) -> VT:
        if isinstance(self.value, Uncached):
            self.value = super().__get__(instance, owner)

        return self.value

    def __set__(self, instance: OT, value: VT) -> None:
        self.value = Uncached()

        super().__set__(instance, value)

    def __delete__(self, instance: OT) -> None:
        self.value = Uncached()

        super().__delete__(instance)


cached_property = CachedProperty
cached_class_property = CachedClassProperty
cached_static_property = CachedStaticProperty



from more_properties.cached_property import (
    cached_class_property,
    cached_property,
    cached_static_property,
)
from more_properties.class_property import class_property, static_property
from more_properties.property import property

__all__ = [
    "property",
    "class_property",
    "classproperty",
    "static_property",
    "staticproperty",
    "cached_property",
    "cached_class_property",
    "cached_static_property",
]

__version__ = "1.0.0"

# Providing aliases for consistency with classmethod and staticmethod
classproperty = class_property
staticproperty = static_property



from typing import Callable, Generic, Optional, Type, TypeVar

__all__ = ["Getable", "Getter", "Setter", "Deleter"]

OT = TypeVar("OT", contravariant=True)  # Owner Type
VT_co = TypeVar("VT_co", covariant=True)  # Covariant Value Type
VT_contra = TypeVar("VT_contra", contravariant=True)  # Contravariant Value Type

try:
    from typing import Protocol
except ImportError:
    Protocol = Generic


class Getable(Protocol[OT, VT_co]):
    def __get__(self, instance: Optional[OT], owner: Type[OT]) -> VT_co:
        ...


Getter = Getable[OT, Callable[[], VT_co]]
Setter = Getable[OT, Callable[[VT_contra], None]]
Deleter = Getable[OT, Callable[[], None]]
